"""
inspircd.py: InspIRCd 2.x protocol module for PyLink.
"""

import time
import threading

from pylinkirc import utils, conf
from pylinkirc.classes import *
from pylinkirc.log import log
from pylinkirc.protocols.ts6_common import *

class InspIRCdProtocol(TS6BaseProtocol):

    S2S_BUFSIZE = 0  # InspIRCd allows infinitely long S2S messages, so set bufsize to infinite

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.protocol_caps |= {'slash-in-nicks', 'slash-in-hosts', 'underscore-in-hosts'}

        # Set our case mapping (rfc1459 maps "\" and "|" together, for example).
        self.casemapping = 'rfc1459'

        # Raw commands sent from servers vary from protocol to protocol. Here, we map
        # non-standard names to our hook handlers, so command handlers' outputs
        # are called with the right hooks.
        self.hook_map = {'FJOIN': 'JOIN', 'RSQUIT': 'SQUIT', 'FMODE': 'MODE',
                    'FTOPIC': 'TOPIC', 'OPERTYPE': 'MODE', 'FHOST': 'CHGHOST',
                    'FIDENT': 'CHGIDENT', 'FNAME': 'CHGNAME', 'SVSTOPIC': 'TOPIC',
                    'SAKICK': 'KICK'}

        self.min_proto_ver = 1202
        self.proto_ver = 1202
        self.max_proto_ver = 1202  # Anything above should warn (not officially supported)

        # Track the modules supported by the uplink.
        self._modsupport = set()

        # Settable by plugins (e.g. relay) as needed, used to work around +j being triggered
        # by bursting users.
        self._endburst_delay = 0

    ### Outgoing commands

    def spawn_client(self, nick, ident='null', host='null', realhost=None, modes=set(),
            server=None, ip='0.0.0.0', realname=None, ts=None, opertype='IRC Operator',
            manipulatable=False):
        """
        Spawns a new client with the given options.

        Note: No nick collision / valid nickname checks are done here; it is
        up to plugins to make sure they don't introduce anything invalid.
        """
        server = server or self.sid

        if not self.is_internal_server(server):
            raise ValueError('Server %r is not a PyLink server!' % server)

        uid = self.uidgen[server].next_uid()

        ts = ts or int(time.time())
        realname = realname or conf.conf['pylink']['realname']
        realhost = realhost or host
        raw_modes = self.join_modes(modes)
        u = self.users[uid] = User(self, nick, ts, uid, server, ident=ident, host=host,
                                   realname=realname, realhost=realhost, ip=ip,
                                   manipulatable=manipulatable, opertype=opertype)

        self.apply_modes(uid, modes)
        self.servers[server].users.add(uid)

        self._send_with_prefix(server, "UID {uid} {ts} {nick} {realhost} {host} {ident} {ip}"
                               " {ts} {modes} + :{realname}".format(ts=ts, host=host,
                               nick=nick, ident=ident, uid=uid,
                               modes=raw_modes, ip=ip, realname=realname,
                               realhost=realhost))
        if ('o', None) in modes or ('+o', None) in modes:
            self._oper_up(uid, opertype)
        return u

    def join(self, client, channel):
        """Joins a PyLink client to a channel."""
        # InspIRCd doesn't distinguish between burst joins and regular joins,
        # so what we're actually doing here is sending FJOIN from the server,
        # on behalf of the clients that are joining.

        server = self.get_server(client)
        if not self.is_internal_server(server):
            log.error('(%s) Error trying to join %r to %r (no such client exists)', self.name, client, channel)
            raise LookupError('No such PyLink client exists.')

        # Strip out list-modes, they shouldn't be ever sent in FJOIN.
        modes = [m for m in self._channels[channel].modes if m[0] not in self.cmodes['*A']]
        self._send_with_prefix(server, "FJOIN {channel} {ts} {modes} :,{uid}".format(
                ts=self._channels[channel].ts, uid=client, channel=channel,
                modes=self.join_modes(modes)))
        self._channels[channel].users.add(client)
        self.users[client].channels.add(channel)

    def sjoin(self, server, channel, users, ts=None, modes=set()):
        """Sends an SJOIN for a group of users to a channel.

        The sender should always be a Server ID (SID). TS is optional, and defaults
        to the one we've stored in the channel state if not given.
        <users> is a list of (prefix mode, UID) pairs:

        Example uses:
            sjoin('100', '#test', [('', '100AAABBC'), ('qo', 100AAABBB'), ('h', '100AAADDD')])
            sjoin(self.sid, '#test', [('o', self.pseudoclient.uid)])
        """
        server = server or self.sid
        assert users, "sjoin: No users sent?"
        log.debug('(%s) sjoin: got %r for users', self.name, users)

        if not server:
            raise LookupError('No such PyLink client exists.')

        # Strip out list-modes, they shouldn't ever be sent in FJOIN (protocol rules).
        modes = modes or self._channels[channel].modes
        orig_ts = self._channels[channel].ts
        ts = ts or orig_ts

        banmodes = []
        regularmodes = []
        for mode in modes:
            modechar = mode[0][-1]
            if modechar in self.cmodes['*A']:
                # Track bans separately (they are sent as a normal FMODE instead of in FJOIN.
                # However, don't reset bans that have already been set.
                if (modechar, mode[1]) not in self._channels[channel].modes:
                    banmodes.append(mode)
            else:
                regularmodes.append(mode)

        uids = []
        changedmodes = set(modes)
        namelist = []

        # We take <users> as a list of (prefixmodes, uid) pairs.
        for userpair in users:
            assert len(userpair) == 2, "Incorrect format of userpair: %r" % userpair
            prefixes, user = userpair
            namelist.append(','.join(userpair))
            uids.append(user)
            for m in prefixes:
                changedmodes.add(('+%s' % m, user))
            try:
                self.users[user].channels.add(channel)
            except KeyError:  # Not initialized yet?
                log.debug("(%s) sjoin: KeyError trying to add %r to %r's channel list?", self.name, channel, user)

        namelist = ' '.join(namelist)
        self._send_with_prefix(server, "FJOIN {channel} {ts} {modes} :{users}".format(
                ts=ts, users=namelist, channel=channel,
                modes=self.join_modes(modes)))
        self._channels[channel].users.update(uids)

        if banmodes:
            # Burst ban modes if there are any.
            # <- :1ML FMODE #test 1461201525 +bb *!*@bad.user *!*@rly.bad.user
            self._send_with_prefix(server, "FMODE {channel} {ts} {modes} ".format(
                ts=ts, channel=channel, modes=self.join_modes(banmodes)))

        self.updateTS(server, channel, ts, changedmodes)

    def _oper_up(self, target, opertype=None):
        """Opers a client up (internal function specific to InspIRCd).

        This should be called whenever user mode +o is set on anyone, because
        InspIRCd requires a special command (OPERTYPE) to be sent in order to
        recognize ANY non-burst oper ups.

        Plugins don't have to call this function themselves, but they can
        set the opertype attribute of an User object (in self.users),
        and the change will be reflected here."""
        userobj = self.users[target]
        try:
            otype = opertype or userobj.opertype or 'IRC Operator'
        except AttributeError:
            log.debug('(%s) opertype field for %s (%s) isn\'t filled yet!',
                      self.name, target, userobj.nick)
            # whatever, this is non-standard anyways.
            otype = 'IRC Operator'
        assert otype, "Tried to send an empty OPERTYPE!"
        log.debug('(%s) Sending OPERTYPE from %s to oper them up.',
                  self.name, target)
        userobj.opertype = otype

        # InspIRCd 2.x uses _ in OPERTYPE to denote spaces, while InspIRCd 3.x does not. This is not
        # backwards compatible: spaces in InspIRCd 2.x will cause the oper type to get cut off at
        # the first word, while underscores in InspIRCd 3.x are shown literally as _.
        # We can do the underscore fixing based on the version of our uplink:
        if self.remote_proto_ver < 1205:
            otype = otype.replace(" ", "_")
        else:
            otype = ':' + otype

        self._send_with_prefix(target, 'OPERTYPE %s' % otype)

    def mode(self, numeric, target, modes, ts=None):
        """Sends mode changes from a PyLink client/server."""
        # -> :9PYAAAAAA FMODE #pylink 1433653951 +os 9PYAAAAAA
        # -> :9PYAAAAAA MODE 9PYAAAAAA -i+w

        if (not self.is_internal_client(numeric)) and \
                (not self.is_internal_server(numeric)):
            raise LookupError('No such PyLink client/server exists.')

        if ('+o', None) in modes and not self.is_channel(target):
            # https://github.com/inspircd/inspircd/blob/master/src/modules/m_spanningtree/opertype.cpp#L26-L28
            # Servers need a special command to set umode +o on people.
            self._oper_up(target)

        self.apply_modes(target, modes)
        joinedmodes = self.join_modes(modes)

        if self.is_channel(target):
            ts = ts or self._channels[target].ts
            self._send_with_prefix(numeric, 'FMODE %s %s %s' % (target, ts, joinedmodes))
        else:
            self._send_with_prefix(numeric, 'MODE %s %s' % (target, joinedmodes))

    def kill(self, numeric, target, reason):
        """Sends a kill from a PyLink client/server."""
        if (not self.is_internal_client(numeric)) and \
                (not self.is_internal_server(numeric)):
            raise LookupError('No such PyLink client/server exists.')

        # InspIRCd will show the raw kill message sent from our server as the quit message.
        # So, make the kill look actually like a kill instead of someone quitting with
        # an arbitrary message.
        if numeric in self.servers:
            sourcenick = self.servers[numeric].name
        else:
            sourcenick = self.users[numeric].nick

        self._send_with_prefix(numeric, 'KILL %s :Killed (%s (%s))' % (target, sourcenick, reason))

        self._remove_client(target)

    def topic_burst(self, numeric, target, text):
        """Sends a topic change from a PyLink server. This is usually used on burst."""
        if not self.is_internal_server(numeric):
            raise LookupError('No such PyLink server exists.')
        ts = int(time.time())
        servername = self.servers[numeric].name
        self._send_with_prefix(numeric, 'FTOPIC %s %s %s :%s' % (target, ts, servername, text))
        self._channels[target].topic = text
        self._channels[target].topicset = True

    def knock(self, numeric, target, text):
        """Sends a KNOCK from a PyLink client."""
        if not self.is_internal_client(numeric):
            raise LookupError('No such PyLink client exists.')
        self._send_with_prefix(numeric, 'ENCAP * KNOCK %s :%s' % (target, text))

    def update_client(self, target, field, text):
        """Updates the ident, host, or realname of any connected client."""
        field = field.upper()

        if field not in ('IDENT', 'HOST', 'REALNAME', 'GECOS'):
            raise NotImplementedError("Changing field %r of a client is "
                                      "unsupported by this protocol." % field)

        if self.is_internal_client(target):
            # It is one of our clients, use FIDENT/HOST/NAME.
            if field == 'IDENT':
                self.users[target].ident = text
                self._send_with_prefix(target, 'FIDENT %s' % text)
            elif field == 'HOST':
                self.users[target].host = text
                self._send_with_prefix(target, 'FHOST %s' % text)
            elif field in ('REALNAME', 'GECOS'):
                self.users[target].realname = text
                self._send_with_prefix(target, 'FNAME :%s' % text)
        else:
            # It is a client on another server, use CHGIDENT/HOST/NAME.
            if field == 'IDENT':
                if 'm_chgident.so' not in self._modsupport:
                    raise NotImplementedError('Cannot change idents as m_chgident.so is not loaded')

                self.users[target].ident = text
                self._send_with_prefix(self.sid, 'CHGIDENT %s %s' % (target, text))

                # Send hook payloads for other plugins to listen to.
                self.call_hooks([self.sid, 'CHGIDENT',
                                   {'target': target, 'newident': text}])
            elif field == 'HOST':
                if 'm_chghost.so' not in self._modsupport:
                    raise NotImplementedError('Cannot change hosts as m_chghost.so is not loaded')

                self.users[target].host = text
                self._send_with_prefix(self.sid, 'CHGHOST %s %s' % (target, text))

                self.call_hooks([self.sid, 'CHGHOST',
                                   {'target': target, 'newhost': text}])

            elif field in ('REALNAME', 'GECOS'):
                if 'm_chgname.so' not in self._modsupport:
                    raise NotImplementedError('Cannot change real names as m_chgname.so is not loaded')

                self.users[target].realname = text
                self._send_with_prefix(self.sid, 'CHGNAME %s :%s' % (target, text))

                self.call_hooks([self.sid, 'CHGNAME',
                                   {'target': target, 'newgecos': text}])

    def numeric(self, source, numeric, target, text):
        """Sends raw numerics from a server to a remote client."""
        # InspIRCd 2.0 syntax (undocumented):
        # Essentially what this does is push the raw numeric text after the first ":" towards the
        # given user.
        # <- :70M PUSH 0ALAAAAAA ::midnight.vpn 422 PyLink-devel :Message of the day file is missing.

        # Note: InspIRCd 2.2 uses a new NUM command in this format:
        # :<sid> NUM <numeric source sid> <target uuid> <3 digit number> <params>
        # Take this into consideration if we ever target InspIRCd 2.2, even though m_spanningtree
        # does provide backwards compatibility for commands like this. -GLolol
        self._send_with_prefix(self.sid, 'PUSH %s ::%s %s %s %s' % (target, source, numeric, target, text))

    def away(self, source, text):
        """Sends an AWAY message from a PyLink client. <text> can be an empty string
        to unset AWAY status."""
        if text:
            self._send_with_prefix(source, 'AWAY %s :%s' % (int(time.time()), text))
        else:
            self._send_with_prefix(source, 'AWAY')
        self.users[source].away = text

    def spawn_server(self, name, sid=None, uplink=None, desc=None):
        """
        Spawns a server off a PyLink server. desc (server description)
        defaults to the one in the config. uplink defaults to the main PyLink
        server, and sid (the server ID) is automatically generated if not
        given.

        Endburst delay can be tweaked by setting the _endburst_delay variable
        to a positive value before calling spawn_server(). This can be used to
        prevent PyLink bursts from filling up snomasks and triggering InspIRCd +j.
        """
        # -> :0AL SERVER test.server * 1 0AM :some silly pseudoserver
        uplink = uplink or self.sid
        name = name.lower()

        # "desc" defaults to the configured server description.
        desc = desc or self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']

        if sid is None:  # No sid given; generate one!
            sid = self.sidgen.next_sid()

        assert len(sid) == 3, "Incorrect SID length"
        if sid in self.servers:
            raise ValueError('A server with SID %r already exists!' % sid)

        for server in self.servers.values():
            if name == server.name:
                raise ValueError('A server named %r already exists!' % name)

        if not self.is_internal_server(uplink):
            raise ValueError('Server %r is not a PyLink server!' % uplink)

        if not self.is_server_name(name):
            raise ValueError('Invalid server name %r' % name)

        self.servers[sid] = Server(self, uplink, name, internal=True, desc=desc)
        self._send_with_prefix(uplink, 'SERVER %s * %s %s :%s' % (name, self.servers[sid].hopcount, sid, desc))

        # Endburst delay clutter

        def endburstf():
            # Delay ENDBURST by X seconds if requested.
            if self._aborted.wait(self._endburst_delay):
                # We managed to catch the abort flag before sending ENDBURST, so break
                log.debug('(%s) stopping endburstf() for %s as aborted was set', self.name, sid)
                return
            self._send_with_prefix(sid, 'ENDBURST')

        if self._endburst_delay:
            t = threading.Thread(target=endburstf, name="protocols/inspircd delayed ENDBURST thread for %s@%s" % (sid, self.name))
            t.daemon = True
            t.start()
        else:  # Else, send burst immediately
            self._send_with_prefix(sid, 'ENDBURST')
        return sid

    def set_server_ban(self, source, duration, user='*', host='*', reason='User banned'):
        """
        Sets a server ban.
        """
        # <- :70M ADDLINE G *@10.9.8.7 midnight.local 1433704565 0 :gib reason pls kthx

        assert not (user == host == '*'), "Refusing to set ridiculous ban on *@*"

        # Per https://wiki.inspircd.org/InspIRCd_Spanning_Tree_1.2/ADDLINE, the setter argument can
        # be a max of 64 characters
        self._send_with_prefix(source, 'ADDLINE G %s@%s %s %s %s :%s' % (user, host, self.get_friendly_name(source)[:64],
                                                                         int(time.time()), duration, reason))

    ### Core / command handlers

    def _post_disconnect(self):
        super()._post_disconnect()
        log.debug('(%s) _post_disconnect: clearing _modsupport entries. Last: %s', self.name, self._modsupport)
        self._modsupport.clear()

    def post_connect(self):
        """Initializes a connection to a server."""
        ts = self.start_ts

        f = self.send
        f('CAPAB START %s' % self.proto_ver)
        f('CAPAB CAPABILITIES :PROTOCOL=%s' % self.proto_ver)
        f('CAPAB END')

        host = self.serverdata["hostname"]
        f('SERVER {host} {Pass} 0 {sid} :{sdesc}'.format(host=host,
          Pass=self.serverdata["sendpass"], sid=self.sid,
          sdesc=self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']))

        self._send_with_prefix(self.sid, 'BURST %s' % ts)
        # InspIRCd sends VERSION data on link, instead of whenever requested by a client.
        self._send_with_prefix(self.sid, 'VERSION :%s' % self.version())
        self._send_with_prefix(self.sid, 'ENDBURST')

        # Extban definitions
        self.extbans_acting = {'quiet': 'm:', 'ban_nonick': 'N:', 'ban_blockcolor': 'c:',
                               'ban_partmsgs': 'p:', 'ban_invites': 'A:', 'ban_blockcaps': 'B:',
                               'ban_noctcp': 'C:', 'ban_nokicks': 'Q:', 'ban_stripcolor': 'S:',
                               'ban_nonotice': 'T:'}
        self.extbans_matching = {'ban_inchannel': 'j:', 'ban_realname': 'r:', 'ban_server': 's:',
                                 'ban_certfp': 'z:', 'ban_opertype': 'O:', 'ban_account': 'R:',
                                 # Note: InspIRCd /helpop refers to this as an acting extban, but
                                 # it actually behaves as a matching one...
                                 'ban_unregistered_matching': 'U:'}

    def handle_capab(self, source, command, args):
        """
        Handles the CAPAB command, used for capability negotiation with our
        uplink.
        """
        # 6 CAPAB commands are usually sent on connect: CAPAB START, MODULES,
        # MODSUPPORT, CHANMODES, USERMODES, and CAPABILITIES.
        # The only ones of interest to us are CHANMODES, USERMODES,
        # CAPABILITIES, and MODSUPPORT.

        if args[0] == 'CHANMODES':
            # <- CAPAB CHANMODES :admin=&a allowinvite=A autoop=w ban=b
            # banexception=e blockcolor=c c_registered=r exemptchanops=X
            # filter=g flood=f halfop=%h history=H invex=I inviteonly=i
            # joinflood=j key=k kicknorejoin=J limit=l moderated=m nickflood=F
            # noctcp=C noextmsg=n nokick=Q noknock=K nonick=N nonotice=T
            # official-join=!Y op=@o operonly=O opmoderated=U owner=~q
            # permanent=P private=p redirect=L reginvite=R regmoderated=M
            # secret=s sslonly=z stripcolor=S topiclock=t voice=+v

            # Named modes are essential for a cross-protocol IRC service. We
            # can use InspIRCd as a model here and assign a similar mode map to
            # our cmodes list.
            for modepair in args[-1].split():
                name, char = modepair.split('=')

                # Strip c_ prefixes to be consistent with other protocols.
                name = name.lstrip('c_')

                if name == 'reginvite':  # Reginvite? That's an odd name.
                    name = 'regonly'

                if name == 'founder':  # Channel mode +q
                    # Founder, owner; same thing. m_customprefix allows you to
                    # name it anything you like. The former is config default,
                    # but I personally prefer the latter.
                    name = 'owner'

                if name in ('repeat', 'kicknorejoin'):
                    # Suffix modes using inspircd-specific arguments so that it can
                    # be safely relayed.
                    name += '_insp'

                # We don't care about mode prefixes; just the mode char.
                self.cmodes[name] = char[-1]


        elif args[0] == 'USERMODES':
            # <- CAPAB USERMODES :bot=B callerid=g cloak=x deaf_commonchan=c
            # helpop=h hidechans=I hideoper=H invisible=i oper=o regdeaf=R
            # servprotect=k showwhois=W snomask=s u_registered=r u_stripcolor=S
            # wallops=w

            # Ditto above.
            for modepair in args[-1].split():
                name, char = modepair.split('=')
                # Strip u_ prefixes to be consistent with other protocols.
                name = name.lstrip('u_')
                self.umodes[name] = char

        elif args[0] == 'CAPABILITIES':
            # <- CAPAB CAPABILITIES :NICKMAX=21 CHANMAX=64 MAXMODES=20
            # IDENTMAX=11 MAXQUIT=255 MAXTOPIC=307 MAXKICK=255 MAXGECOS=128
            # MAXAWAY=200 IP6SUPPORT=1 PROTOCOL=1202 PREFIX=(Yqaohv)!~&@%+
            # CHANMODES=IXbegw,k,FHJLfjl,ACKMNOPQRSTUcimnprstz
            # USERMODES=,,s,BHIRSWcghikorwx GLOBOPS=1 SVSPART=1

            # First, turn the arguments into a dict
            caps = self.parse_isupport(args[-1])
            log.debug("(%s) capabilities list: %s", self.name, caps)

            # Check the protocol version
            self.remote_proto_ver = protocol_version = int(caps['PROTOCOL'])

            if protocol_version < self.min_proto_ver:
                raise ProtocolError("Remote protocol version is too old! "
                                    "At least %s (InspIRCd 2.0.x) is "
                                    "needed. (got %s)" % (self.min_proto_ver,
                                                          protocol_version))
            elif protocol_version > self.max_proto_ver:
                log.warning("(%s) PyLink support for InspIRCd 2.2+ is experimental, "
                            "and should not be relied upon for anything important.",
                            self.name)

            # Store the max nick and channel lengths
            self.maxnicklen = int(caps['NICKMAX'])
            self.maxchanlen = int(caps['CHANMAX'])

            # Modes are divided into A, B, C, and D classes
            # See http://www.irc.org/tech_docs/005.html

            # FIXME: Find a neater way to assign/store this.
            self.cmodes['*A'], self.cmodes['*B'], self.cmodes['*C'], self.cmodes['*D'] \
                = caps['CHANMODES'].split(',')
            self.umodes['*A'], self.umodes['*B'], self.umodes['*C'], self.umodes['*D'] \
                = caps['USERMODES'].split(',')

            # Separate the prefixes field (e.g. "(Yqaohv)!~&@%+") into a
            # dict mapping mode characters to mode prefixes.
            self.prefixmodes = self.parse_isupport_prefixes(caps['PREFIX'])
            log.debug('(%s) self.prefixmodes set to %r', self.name,
                      self.prefixmodes)

        elif args[0] == 'MODSUPPORT':
            # <- CAPAB MODSUPPORT :m_alltime.so m_check.so m_chghost.so m_chgident.so m_chgname.so m_fullversion.so m_gecosban.so m_knock.so m_muteban.so m_nicklock.so m_nopartmsg.so m_opmoderated.so m_sajoin.so m_sanick.so m_sapart.so m_serverban.so m_services_account.so m_showwhois.so m_silence.so m_swhois.so m_uninvite.so m_watch.so
            self._modsupport |= set(args[-1].split())

    def handle_ping(self, source, command, args):
        """Handles incoming PING commands, so we don't time out."""
        # <- :70M PING 70M 0AL
        # -> :0AL PONG 0AL 70M
        if len(args) >= 2:
            self._send_with_prefix(args[1], 'PONG %s %s' % (args[1], source), queue=False)
        else:
            self._send_with_prefix(self.sid, 'PONG %s' % source, queue=False)

    def handle_fjoin(self, servernumeric, command, args):
        """Handles incoming FJOIN commands (InspIRCd equivalent of JOIN/SJOIN)."""
        # :70M FJOIN #chat 1423790411 +AFPfjnt 6:5 7:5 9:5 :o,1SRAABIT4 v,1IOAAF53R <...>
        channel = args[0]
        chandata = self._channels[channel].deepcopy()
        # InspIRCd sends each channel's users in the form of 'modeprefix(es),UID'
        userlist = args[-1].split()

        modestring = args[2:-1] or args[2]
        parsedmodes = self.parse_modes(channel, modestring)
        namelist = []

        # Keep track of other modes that are added due to prefix modes being joined too.
        changedmodes = set(parsedmodes)

        for user in userlist:
            modeprefix, user = user.split(',', 1)

            # Don't crash when we get an invalid UID.
            if user not in self.users:
                log.debug('(%s) handle_fjoin: tried to introduce user %s not in our user list, ignoring...',
                          self.name, user)
                continue

            namelist.append(user)
            self.users[user].channels.add(channel)

            # Only save mode changes if the remote has lower TS than us.
            changedmodes |= {('+%s' % mode, user) for mode in modeprefix}

            self._channels[channel].users.add(user)

        # Statekeeping with timestamps. Note: some service packages (Anope 1.8) send a trailing
        # 'd' after the timestamp, which we should strip out to prevent int() from erroring.
        # This is technically valid in InspIRCd S2S because atoi() ignores non-digit characters,
        # but it's strange behaviour either way...
        # <- :3AX FJOIN #monitor 1485462109d + :,3AXAAAAAK
        their_ts = int(''.join(char for char in args[1] if char.isdigit()))

        our_ts = self._channels[channel].ts
        self.updateTS(servernumeric, channel, their_ts, changedmodes)

        return {'channel': channel, 'users': namelist, 'modes': parsedmodes, 'ts': their_ts,
                'channeldata': chandata}

    def handle_uid(self, numeric, command, args):
        """Handles incoming UID commands (user introduction)."""
        # :70M UID 70MAAAAAB 1429934638 GL 0::1 hidden-7j810p.9mdf.lrek.0000.0000.IP gl 0::1 1429934638 +Wioswx +ACGKNOQXacfgklnoqvx :realname
        uid, ts, nick, realhost, host, ident, ip = args[0:7]

        ts = int(ts)

        self._check_nick_collision(nick)
        realname = args[-1]
        self.users[uid] = userobj = User(self, nick, ts, uid, numeric, ident, host, realname, realhost, ip)

        parsedmodes = self.parse_modes(uid, [args[8], args[9]])
        self.apply_modes(uid, parsedmodes)

        self._check_oper_status_change(uid, parsedmodes)

        self.servers[numeric].users.add(uid)
        return {'uid': uid, 'ts': ts, 'nick': nick, 'realhost': realhost, 'host': host, 'ident': ident, 'ip': ip}

    def handle_server(self, numeric, command, args):
        """Handles incoming SERVER commands (introduction of servers)."""

        # Initial SERVER command on connect.
        if self.uplink is None:
            # <- SERVER whatever.net abcdefgh 0 10X :some server description
            servername = args[0].lower()
            numeric = args[3]

            if args[1] != self.serverdata['recvpass']:
                 # Check if recvpass is correct
                 raise ProtocolError('recvpass from uplink server %s does not match configuration!' % servername)

            sdesc = args[-1]
            self.servers[numeric] = Server(self, None, servername, desc=sdesc)
            self.uplink = numeric
            return

        # Other server introductions.
        # <- :00A SERVER test.server * 1 00C :testing raw message syntax
        servername = args[0].lower()
        sid = args[3]
        sdesc = args[-1]
        self.servers[sid] = Server(self, numeric, servername, desc=sdesc)

        return {'name': servername, 'sid': args[3], 'text': sdesc}

    def handle_fmode(self, numeric, command, args):
        """Handles the FMODE command, used for channel mode changes."""
        # <- :70MAAAAAA FMODE #chat 1433653462 +hhT 70MAAAAAA 70MAAAAAD
        channel = args[0]
        oldobj = self._channels[channel].deepcopy()
        modes = args[2:]
        changedmodes = self.parse_modes(channel, modes)
        self.apply_modes(channel, changedmodes)
        ts = int(args[1])
        return {'target': channel, 'modes': changedmodes, 'ts': ts,
                'channeldata': oldobj}

    def handle_idle(self, source, command, args):
        """
        Handles the IDLE command, sent between servers in remote WHOIS queries.
        """
        # <- :70MAAAAAA IDLE 1MLAAAAIG
        # -> :1MLAAAAIG IDLE 70MAAAAAA 1433036797 319

        if self.serverdata.get('force_whois_extensions', True):
            return {'target': args[0], 'parse_as': 'WHOIS'}

        # Allow hiding the startup time if set to do so (if both idle and signon time is 0, InspIRCd omits
        # showing this line).
        target = args[0]
        start_time = self.start_ts if (conf.conf['pylink'].get('whois_show_startup_time', True) and
                                       self.get_service_bot(target)) else 0

        # First arg = source, second = signon time, third = idle time
        self._send_with_prefix(target, 'IDLE %s %s 0' % (source, start_time))

    def handle_ftopic(self, numeric, command, args):
        """Handles incoming FTOPIC (sets topic on burst)."""
        # <- :70M FTOPIC #channel 1434510754 GLo|o|!GLolol@escape.the.dreamland.ca :Some channel topic
        channel = args[0]
        ts = args[1]
        setter = args[2]
        topic = args[-1]
        self._channels[channel].topic = topic
        self._channels[channel].topicset = True
        return {'channel': channel, 'setter': setter, 'ts': ts, 'text': topic}

    # SVSTOPIC is used by InspIRCd module m_topiclock - its arguments are the same as FTOPIC
    handle_svstopic = handle_ftopic

    def handle_opertype(self, target, command, args):
        """Handles incoming OPERTYPE, which is used to denote an oper up.

        This calls the internal hook CLIENT_OPERED, sets the internal
        opertype of the client, and assumes setting user mode +o on the caller."""
        # This is used by InspIRCd to denote an oper up; there is no MODE
        # command sent for it.
        # <- :70MAAAAAB OPERTYPE Network_Owner
        # Replace escaped _ in opertypes with spaces for InspIRCd 2.0.
        opertype = args[0].replace("_", " ")

        # Set umode +o on the target.
        omode = [('+o', None)]
        self.apply_modes(target, omode)

        # Call the CLIENT_OPERED hook that protocols use. The MODE hook
        # payload is returned below.
        self.call_hooks([target, 'CLIENT_OPERED', {'text': opertype}])
        return {'target': target, 'modes': omode}

    def handle_fident(self, numeric, command, args):
        """Handles FIDENT, used for denoting ident changes."""
        # <- :70MAAAAAB FIDENT test
        self.users[numeric].ident = newident = args[0]
        return {'target': numeric, 'newident': newident}

    def handle_fhost(self, numeric, command, args):
        """Handles FHOST, used for denoting hostname changes."""
        # <- :70MAAAAAB FHOST some.host
        self.users[numeric].host = newhost = args[0]
        return {'target': numeric, 'newhost': newhost}

    def handle_fname(self, numeric, command, args):
        """Handles FNAME, used for denoting real name/gecos changes."""
        # <- :70MAAAAAB FNAME :afdsafasf
        self.users[numeric].realname = newgecos = args[0]
        return {'target': numeric, 'newgecos': newgecos}

    def handle_endburst(self, numeric, command, args):
        """ENDBURST handler; sends a hook with empty contents."""
        self.servers[numeric].has_eob = True
        if numeric == self.uplink:
            self.connected.set()
        return {}

    def handle_away(self, numeric, command, args):
        """Handles incoming AWAY messages."""
        # <- :1MLAAAAIG AWAY 1439371390 :Auto-away
        try:
            ts = args[0]
            self.users[numeric].away = text = args[1]
            return {'text': text, 'ts': ts}
        except IndexError:  # User is unsetting away status
            self.users[numeric].away = ''
            return {'text': ''}

    def handle_rsquit(self, numeric, command, args):
        """
        Handles the RSQUIT command, which is sent by opers to SQUIT remote
        servers.
        """
        # <- :1MLAAAAIG RSQUIT :ayy.lmao
        # <- :1MLAAAAIG RSQUIT ayy.lmao :some reason

        # RSQUIT is sent by opers to SQUIT remote servers. However, it differs from
        # a regular SQUIT in that:
        #    1) It takes a server name instead of a SID,
        #    2) Responses have to be be explicitly sent; i.e. The target server has
        #       to agree with splitting the target, and could ignore such requests
        #       entirely.

        # If we receive such a remote SQUIT, just forward it as a regular
        # SQUIT, in order to be consistent with other IRCds which make SQUITs
        # implicit.
        target = self._get_SID(args[0])
        if self.is_internal_server(target):
            # The target has to be one of our servers in order to work...
            uplink = self.servers[target].uplink
            reason = 'Requested by %s' % self.get_hostmask(numeric)
            self._send_with_prefix(uplink, 'SQUIT %s :%s' % (target, reason))
            return self.handle_squit(numeric, 'SQUIT', [target, reason])
        else:
            log.debug("(%s) Got RSQUIT for '%s', which is either invalid or not "
                      "a server of ours!", self.name, args[0])

    def handle_metadata(self, numeric, command, args):
        """
        Handles the METADATA command, used by servers to send metadata (services
        login name, certfp data, etc.) for clients.
        """
        uid = args[0]

        if args[1] == 'accountname' and uid in self.users:
            # <- :00A METADATA 1MLAAAJET accountname :
            # <- :00A METADATA 1MLAAAJET accountname :tester
            # Sets the services login name of the client.

            self.call_hooks([uid, 'CLIENT_SERVICES_LOGIN', {'text': args[-1]}])
        elif args[1] == 'modules' and numeric == self.uplink:
            # Note: only handle METADATA from our uplink; otherwise leaf servers unloading modules
            # while shutting down will corrupt the state.
            # <- :70M METADATA * modules :-m_chghost.so
            # <- :70M METADATA * modules :+m_chghost.so
            for module in args[-1].split():
                if module.startswith('-'):
                    log.debug('(%s) Removing module %s', self.name, module[1:])
                    self._modsupport.discard(module[1:])
                elif module.startswith('+'):
                    log.debug('(%s) Adding module %s', self.name, module[1:])
                    self._modsupport.add(module[1:])
                else:
                    log.warning('(%s) Got unknown METADATA modules string: %r', self.name, args[-1])

    def handle_version(self, numeric, command, args):
        """
        Stub VERSION handler (does nothing) to override the one in ts6_common.
        """

    def handle_sakick(self, source, command, args):
        """Handles forced kicks (SAKICK)."""
        # <- :1MLAAAAAD ENCAP 0AL SAKICK #test 0ALAAAAAB :test
        # ENCAP -> SAKICK args: ['#test', '0ALAAAAAB', 'test']

        target = args[1]
        channel = args[0]
        try:
            reason = args[2]
        except IndexError:
            # Kick reason is optional, strange...
            reason = self.get_friendly_name(source)

        if not self.is_internal_client(target):
            log.warning("(%s) Got SAKICK for client that not one of ours: %s", self.name, target)
            return
        else:
            # Like RSQUIT, SAKICK requires that the receiving server acknowledge that a kick has
            # happened. This comes from the server hosting the target client.
            server = self.get_server(target)

        self.kick(server, channel, target, reason)
        return {'channel': channel, 'target': target, 'text': reason}

    def handle_alltime(self, source, command, args):
        """Handles /ALLTIME requests."""
        # -> :9PYAAAAAA ENCAP * ALLTIME
        # <- :70M PUSH 0ALAAAAAC ::midnight.vpn NOTICE PyLink-devel :System time is 2016-08-13 02:23:06 (1471054986) on midnight.vpn

        # XXX: We override notice() here because that abstraction doesn't allow messages from servers.
        timestring = '%s (%s)' % (time.strftime('%Y-%m-%d %H:%M:%S'), int(time.time()))
        self._send_with_prefix(self.sid, 'NOTICE %s :System time is %s on %s' % (source, timestring, self.hostname()))

Class = InspIRCdProtocol



"""Setup module for PyLink IRC Services."""

import sys

if sys.version_info < (3, 4):
    raise RuntimeError("PyLink requires Python 3.4 or higher.")

try:
    from setuptools import setup, find_packages
except ImportError:
    raise ImportError("Please install Setuptools and try again.")
from codecs import open
import subprocess

with open('VERSION', encoding='utf-8') as f:
    version = f.read().strip()

# Try to fetch the current commit hash from Git.
try:
    real_version = subprocess.check_output(['git', 'describe', '--tags']).decode('utf-8').strip()
except Exception as e:
    print('WARNING: Failed to get Git version from "git describe --tags": %s: %s' % (type(e).__name__, e))
    print("If you're installing from PyPI or a tarball, ignore the above message.")
    real_version = version + '-nogit'

# Write the version to disk.
with open('__init__.py', 'w') as f:
    f.write('# Automatically generated by setup.py\n')
    f.write('__version__ = %r\n' % version)
    f.write('real_version = %r\n' % real_version)

try:
    if sys.version_info >= (3, 5):
        with open('README.md') as f:
            long_description = f.read()
    else:
        # Work around "TypeError: a bytes-like object is required, not 'str'" errors on Python 3.4
        # when the README has Unicode characters (error in distutils.util.rfc822_escape)
        import codecs
        long_description = codecs.open('README.md', encoding='utf-8').read()
except OSError:
    print('WARNING: Failed to read readme, skipping writing long_description')
    long_description = None

setup(
    name='pylinkirc',
    version=version,

    description='PyLink IRC Services',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/jlu5/PyLink',

    # Author details
    author='James Lu',
    author_email='james@overdrivenetworks.com',

    # License
    license='MPL 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Environment :: Console',

        'Operating System :: OS Independent',
        'Operating System :: POSIX',

        'Natural Language :: English',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='IRC services relay',
    install_requires=['pyyaml', 'ircmatch'],

    extras_require={
        'password-hashing': ['passlib'],
        'cron-support': ['psutil'],
        'servprotect': ['expiringdict>=1.1.4'],
    },

    dependency_links=[
        "git+https://github.com/mailgun/expiringdict.git@v1.1.4#egg=expiringdict-1.1.4"
    ],

    # Folders (packages of code)
    packages=['pylinkirc', 'pylinkirc.protocols', 'pylinkirc.plugins', 'pylinkirc.coremods'],

    # Data files
    package_data={
        '': ['example-conf.yml', 'VERSION', 'README.md'],
    },

    package_dir = {'pylinkirc': '.'},

    # Executable scripts
    scripts=["pylink-mkpasswd"],

    entry_points = {
        'console_scripts': ['pylink=pylinkirc.launcher:main'],
    }
)



#!/usr/bin/env python3
"""
PyLink IRC Services launcher.
"""

import os
import sys
import signal
import time
from pylinkirc import world, conf, __version__, real_version

try:
    import psutil
except ImportError:
    psutil = None

args = {}

def _main():
    conf.load_conf(args.config)

    from pylinkirc.log import log
    from pylinkirc import classes, utils, coremods, selectdriver

    # Write and check for an existing PID file unless specifically told not to.
    if not args.no_pid:
        pidfile = '%s.pid' % conf.confname
        pid_exists = False
        pid = None
        if os.path.exists(pidfile):
            try:
                with open(pidfile) as f:
                    pid = int(f.read())
            except OSError:
                log.exception("Could not read PID file %s:", pidfile)
            else:
                pid_exists = True

            if psutil is not None and os.name == 'posix':
                # FIXME: Haven't tested this on other platforms, so not turning it on by default.
                try:
                    proc = psutil.Process(pid)
                except psutil.NoSuchProcess:  # Process doesn't exist!
                    pid_exists = False
                    log.info("Ignoring stale PID %s from PID file %r: no such process exists.", pid, pidfile)
                else:
                    # This PID got reused for something that isn't us?
                    if not any('pylink' in arg.lower() for arg in proc.cmdline()):
                        log.info("Ignoring stale PID %s from PID file %r: process command line %r is not us", pid, pidfile, proc.cmdline())
                        pid_exists = False

        if pid and pid_exists:
            if args.rehash:
                os.kill(pid, signal.SIGUSR1)
                log.info('OK, rehashed PyLink instance %s (config %r)', pid, args.config)
                sys.exit()
            elif args.stop or args.restart:  # Handle --stop and --restart options
                os.kill(pid, signal.SIGTERM)

                log.info("Waiting for PyLink instance %s (config %r) to stop...", pid, args.config)
                while os.path.exists(pidfile):
                    # XXX: this is ugly, but os.waitpid() only works on non-child processes on Windows
                    time.sleep(0.2)
                log.info("Successfully killed PID %s for config %r.", pid, args.config)

                if args.stop:
                    sys.exit()
            else:
                log.error("PID file %r exists; aborting!", pidfile)

                if psutil is None:
                    log.error("If PyLink didn't shut down cleanly last time it ran, or you're upgrading "
                              "from PyLink < 1.1-dev, delete %r and start the server again.", pidfile)
                    if os.name == 'posix':
                        log.error("Alternatively, you can install psutil for Python 3 (pip3 install psutil), "
                                  "which will allow this launcher to detect stale PID files and ignore them.")
                sys.exit(1)

        elif args.stop or args.restart or args.rehash:  # XXX: also repetitive
            # --stop and --restart should take care of stale PIDs.
            if pid:
                world._should_remove_pid = True
                log.error('Cannot stop/rehash PyLink: no process with PID %s exists.', pid)
            else:
                log.error('Cannot stop/rehash PyLink: PID file %r does not exist or cannot be read.', pidfile)
            sys.exit(1)

        world._should_remove_pid = True

    log.info('PyLink %s starting...', __version__)

    world.daemon = args.daemonize
    if args.daemonize:
        if args.no_pid:
            print('ERROR: Combining --no-pid and --daemonize is not supported.')
            sys.exit(1)
        elif os.name != 'posix':
            print('ERROR: Daemonization is not supported outside POSIX systems.')
            sys.exit(1)
        else:
            log.info('Forking into the background.')
            log.removeHandler(world.console_handler)

            # Adapted from https://stackoverflow.com/questions/5975124/
            if os.fork():
                # Fork and exit the parent process.
                os._exit(0)

            os.setsid()  # Decouple from our lovely terminal
            if os.fork():
                # Fork again to prevent starting zombie apocalypses.
                os._exit(0)
    else:
        # For foreground sessions, set the terminal window title.
        # See https://bbs.archlinux.org/viewtopic.php?id=85567 &
        #     https://stackoverflow.com/questions/7387276/
        if os.name == 'nt':
            import ctypes
            ctypes.windll.kernel32.SetConsoleTitleW("PyLink %s" % __version__)
        elif os.name == 'posix':
            sys.stdout.write("\x1b]2;PyLink %s\x07" % __version__)

    if not args.no_pid:
        # Write the PID file only after forking.
        with open(pidfile, 'w') as f:
            f.write(str(os.getpid()))

    # Load configured plugins
    to_load = conf.conf['plugins']
    utils._reset_module_dirs()

    for plugin in to_load:
        try:
            world.plugins[plugin] = pl = utils._load_plugin(plugin)
        except Exception as e:
            log.exception('Failed to load plugin %r: %s: %s', plugin, type(e).__name__, str(e))
        else:
            if hasattr(pl, 'main'):
                log.debug('Calling main() function of plugin %r', pl)
                pl.main()

    # Initialize all the networks one by one
    for network, sdata in conf.conf['servers'].items():
        try:
            protoname = sdata['protocol']
        except (KeyError, TypeError):
            log.error("(%s) Configuration error: No protocol module specified, aborting.", network)
        else:
            # Fetch the correct protocol module.
            try:
                proto = utils._get_protocol_module(protoname)

                # Create and connect the network.
                world.networkobjects[network] = irc = proto.Class(network)
                log.debug('Connecting to network %r', network)
                irc.connect()
            except:
                log.exception('(%s) Failed to connect to network %r, skipping it...',
                              network, network)
                continue

    world.started.set()
    log.info("Loaded plugins: %s", ', '.join(sorted(world.plugins.keys())))
    selectdriver.start()

def main():
    import argparse

    global args

    parser = argparse.ArgumentParser(description='Starts an instance of PyLink IRC Services.')
    parser.add_argument('config', help='specifies the path to the config file (defaults to pylink.yml)', nargs='?', default='pylink.yml')
    parser.add_argument("-v", "--version", help="displays the program version and exits", action='store_true')
    parser.add_argument("-c", "--check-pid", help="no-op; kept for compatibility with PyLink <= 1.2.x", action='store_true')
    parser.add_argument("-n", "--no-pid", help="skips generating and checking PID files", action='store_true')
    parser.add_argument("-r", "--restart", help="restarts the PyLink instance with the given config file", action='store_true')
    parser.add_argument("-s", "--stop", help="stops the PyLink instance with the given config file", action='store_true')
    parser.add_argument("-R", "--rehash", help="rehashes the PyLink instance with the given config file", action='store_true')
    parser.add_argument("-d", "--daemonize", help="[experimental] daemonizes the PyLink instance on POSIX systems", action='store_true')
    parser.add_argument("-t", "--trace", help="traces through running Python code; useful for debugging", action='store_true')
    parser.add_argument('--trace-ignore-mods', help='comma-separated list of extra modules to ignore when tracing', action='store', default='')
    parser.add_argument('--trace-ignore-dirs', help='comma-separated list of extra directories to ignore when tracing', action='store', default='')
    args = parser.parse_args()

    if args.version:  # Display version and exit
        print('PyLink %s (in VCS: %s)' % (__version__, real_version))
        sys.exit()

    # XXX: repetitive
    elif args.no_pid and (args.restart or args.stop or args.rehash):
        print('ERROR: --no-pid cannot be combined with --restart or --stop')
        sys.exit(1)
    elif args.rehash and os.name != 'posix':
        print('ERROR: Rehashing via the command line is not supported outside Unix.')
        sys.exit(1)

    if args.trace:
        import trace
        tracer = trace.Trace(ignoremods=args.trace_ignore_mods.split(','),
                             ignoredirs=args.trace_ignore_dirs.split(','))
        tracer.runctx('_main()', globals=globals(), locals=locals())
    else:
        _main()



"""
automode.py - Provide simple channel ACL management by giving prefix modes to users matching
hostmasks or exttargets.
"""
import collections

from pylinkirc import utils, conf, world, structures
from pylinkirc.log import log
from pylinkirc.coremods import permissions

mydesc = ("The \x02Automode\x02 plugin provides simple channel ACL management by giving prefix modes "
          "to users matching hostmasks or exttargets.")

# Register ourselves as a service.
modebot = utils.register_service("automode", default_nick="Automode", desc=mydesc)
reply = modebot.reply
error = modebot.error

# Databasing variables.
dbname = conf.get_database_name('automode')
datastore = structures.JSONDataStore('automode', dbname, default_db=collections.defaultdict(dict))

db = datastore.store

# The default set of Automode permissions.
default_permissions = {"$ircop": ['automode.manage.relay_owned', 'automode.sync.relay_owned',
                                  'automode.list']}

def _join_db_channels(irc):
    """
    Joins the Automode service client to channels on the current network in its DB.
    """
    if not irc.connected.is_set():
        log.debug('(%s) _join_db_channels: aborting, network not ready yet', irc.name)
        return

    for entry in db:
        netname, channel = entry.split('#', 1)
        channel = '#' + channel
        if netname == irc.name:
            modebot.add_persistent_channel(irc, 'automode', channel)

def main(irc=None):
    """Main function, called during plugin loading."""

    # Load the automode database.
    datastore.load()

    # Register our permissions.
    permissions.add_default_permissions(default_permissions)

    if irc:  # This was a reload.
        for ircobj in world.networkobjects.values():
            _join_db_channels(ircobj)

def die(irc=None):
    """Saves the Automode database and quit."""
    datastore.die()
    permissions.remove_default_permissions(default_permissions)
    utils.unregister_service('automode')

def _check_automode_access(irc, uid, channel, command):
    """Checks the caller's access to Automode."""
    # Automode defines the following permissions, where <command> is either "manage", "list",
    # "sync", "clear", "remotemanage", "remotelist", "remotesync", "remoteclear":
    # - automode.<command> OR automode.<command>.*: ability to <command> automode on all channels.
    # - automode.<command>.relay_owned: ability to <command> automode on channels owned via Relay.
    #   If Relay isn't loaded, this permission check FAILS.
    # - automode.<command>.#channel: ability to <command> automode on the given channel.
    # - automode.savedb: ability to save the automode DB.
    log.debug('(%s) Automode: checking access for %s/%s for %s capability on %s', irc.name, uid,
              irc.get_hostmask(uid), command, channel)

    baseperm = 'automode.%s' % command
    try:
        # First, check the catch all and channel permissions.
        perms = [baseperm, baseperm+'.*', '%s.%s' % (baseperm, channel)]
        return permissions.check_permissions(irc, uid, perms)
    except utils.NotAuthorizedError:
        if not command.startswith('remote'):
            # Relay-based ACL checking only works with local calls.
            log.debug('(%s) Automode: falling back to automode.%s.relay_owned', irc.name, command)
            permissions.check_permissions(irc, uid, [baseperm+'.relay_owned'], also_show=perms)

            relay = world.plugins.get('relay')
            if relay is None:
                raise utils.NotAuthorizedError("You are not authorized to use Automode when Relay is "
                                               "disabled. You are missing one of the following "
                                               "permissions: %s or %s.%s" % (baseperm, baseperm, channel))
            elif (irc.name, channel) not in relay.db:
                raise utils.NotAuthorizedError("The network you are on does not own the relay channel %s." % channel)
            return True
        raise

def match(irc, channel, uids=None):
    """
    Automode matcher engine.
    """
    dbentry = db.get(irc.name+channel)
    if dbentry is None:
        return

    modebot_uid = modebot.uids.get(irc.name)

    # Check every mask defined in the channel ACL.
    outgoing_modes = []

    # If UIDs are given, match those. Otherwise, match all users in the given channel.
    uids = uids or irc.channels[channel].users

    for mask, modes in dbentry.items():
        for uid in uids:
            if irc.match_host(mask, uid):
                # User matched a mask. Filter the mode list given to only those that are valid
                # prefix mode characters.
                outgoing_modes += [('+'+mode, uid) for mode in modes if mode in irc.prefixmodes]
                log.debug("(%s) automode: Filtered mode list of %s to %s (protocol:%s)",
                          irc.name, modes, outgoing_modes, irc.protoname)

    if outgoing_modes:
        # If the Automode bot is missing, send the mode through the PyLink server.
        if modebot_uid not in irc.users:
            modebot_uid = irc.sid

        log.debug("(%s) automode: sending modes from modebot_uid %s",
                  irc.name, modebot_uid)

        irc.mode(modebot_uid, channel, outgoing_modes)

        # Create a hook payload to support plugins like relay.
        irc.call_hooks([modebot_uid, 'AUTOMODE_MODE',
                      {'target': channel, 'modes': outgoing_modes, 'parse_as': 'MODE'}])

def handle_endburst(irc, source, command, args):
    """ENDBURST hook handler - used to join the Automode service to channels where it has entries."""
    if source == irc.uplink:
        _join_db_channels(irc)
utils.add_hook(handle_endburst, 'ENDBURST')

def handle_join(irc, source, command, args):
    """
    Automode JOIN listener. This sets modes accordingly if the person joining matches a mask in the
    ACL.
    """
    channel = irc.to_lower(args['channel'])
    match(irc, channel, args['users'])

utils.add_hook(handle_join, 'JOIN')
utils.add_hook(handle_join, 'PYLINK_RELAY_JOIN')  # Handle the relay version of join
utils.add_hook(handle_join, 'PYLINK_SERVICE_JOIN')  # And the version for service bots

def handle_services_login(irc, source, command, args):
    """
    Handles services login change, to trigger Automode matching.
    """
    for channel in irc.users[source].channels:
        # Look at all the users' channels for any possible changes.
        match(irc, channel, [source])

utils.add_hook(handle_services_login, 'CLIENT_SERVICES_LOGIN')
utils.add_hook(handle_services_login, 'PYLINK_RELAY_SERVICES_LOGIN')

def _get_channel_pair(irc, source, chanpair, perm=None):
    """
    Fetches the network and channel given a channel pair,
    also optionally checking the caller's permissions.
    """
    log.debug('(%s) Looking up chanpair %s', irc.name, chanpair)
    try:
        network, channel = chanpair.split('#', 1)
    except ValueError:
        raise ValueError("Invalid channel pair %r" % chanpair)
    channel = '#' + channel
    channel = irc.to_lower(channel)

    if not irc.is_channel(channel):
        raise ValueError("Invalid channel name %s." % channel)

    if network:
        ircobj = world.networkobjects.get(network)
    else:
        ircobj = irc

    if not ircobj:
        raise ValueError("Unknown network %s" % network)

    if perm is not None:
        # Only check for permissions if we're told to and the irc object exists.
        if ircobj.name != irc.name:
            perm = 'remote' + perm

        _check_automode_access(irc, source, channel, perm)

    return (ircobj, channel)

def setacc(irc, source, args):
    """<channel/chanpair> <mask> <mode list>

    Assigns the given prefix mode characters to the given mask for the channel given. Extended targets are supported for masks - use this to your advantage!

    Channel pairs are also supported (for operations on remote channels), using the form "network#channel".


    Examples:

    \x02SETACC #channel *!*@localhost ohv

    \x02SETACC #channel $account v

    \x02SETACC othernet#channel $ircop:Network?Administrator qo

    \x02SETACC #staffchan $channel:#mainchan:op o
    """

    try:
        chanpair, mask, modes = args
    except ValueError:
        error(irc, "Invalid arguments given. Needs 3: channel, mask, mode list.")
        return
    else:
        ircobj, channel = _get_channel_pair(irc, source, chanpair, perm='manage')

    # Database entries for any network+channel pair are automatically created using
    # defaultdict. Note: string keys are used here instead of tuples so they can be
    # exported easily as JSON.
    dbentry = db[ircobj.name+channel]

    modes = modes.lstrip('+')  # remove extraneous leading +'s
    dbentry[mask] = modes
    log.info('(%s) %s set modes +%s for %s on %s', ircobj.name, irc.get_hostmask(source), modes, mask, channel)
    reply(irc, "Done. \x02%s\x02 now has modes \x02+%s\x02 in \x02%s\x02." % (mask, modes, channel))

    # Join the Automode bot to the channel persistently.
    modebot.add_persistent_channel(ircobj, 'automode', channel)

modebot.add_cmd(setacc, aliases=('setaccess', 'set'), featured=True)

def delacc(irc, source, args):
    """<channel/chanpair> <mask or range string>

    Removes the Automode entry for the given mask or range string, if they exist.

    Range strings are indices (entry numbers) or ranges of them joined together with commas: e.g.
    "1", "2-10", "1,3,5-8". Entry numbers are shown by LISTACC.
    """
    try:
        chanpair, mask = args
    except ValueError:
        error(irc, "Invalid arguments given. Needs 2: channel, mask")
        return
    else:
        ircobj, channel = _get_channel_pair(irc, source, chanpair, perm='manage')

    dbentry = db.get(ircobj.name+channel)

    if dbentry is None:
        error(irc, "No Automode access entries exist for \x02%s\x02." % channel)
        return

    if mask in dbentry:
        del dbentry[mask]
        log.info('(%s) %s removed modes for %s on %s', ircobj.name, irc.get_hostmask(source), mask, channel)
        reply(irc, "Done. Removed the Automode access entry for \x02%s\x02 in \x02%s\x02." % (mask, channel))
    else:
        # Treat the mask as a range string.
        try:
            new_keys = utils.remove_range(mask, sorted(dbentry.keys()))
        except ValueError:
            error(irc, "No Automode access entry for \x02%s\x02 exists in \x02%s\x02." % (mask, channel))
            return

        # XXX: Automode entries are actually unordered: what we're actually doing is sorting the keys
        # by name into a list, running remove_range on that, and removing the difference.
        removed = []
        source_host = irc.get_hostmask(source)
        for mask_entry in dbentry.copy():
            if mask_entry not in new_keys:
                del dbentry[mask_entry]
                log.info('(%s) %s removed modes for %s on %s', ircobj.name, source_host, mask_entry, channel)
                removed.append(mask_entry)

        reply(irc, 'Done. Removed \x02%d\x02 entries on \x02%s\x02: %s' % (len(removed), channel, ', '.join(removed)))

    # Remove channels if no more entries are left.
    if not dbentry:
        log.debug("Automode: purging empty channel pair %s/%s", ircobj.name, channel)
        del db[ircobj.name+channel]
        modebot.remove_persistent_channel(ircobj, 'automode', channel)

modebot.add_cmd(delacc, aliases=('delaccess', 'del'), featured=True)

def listacc(irc, source, args):
    """<channel/chanpair>

    Lists all Automode entries for the given channel."""
    try:
        chanpair = args[0]
    except IndexError:
        error(irc, "Invalid arguments given. Needs 1: channel.")
        return
    else:
        ircobj, channel = _get_channel_pair(irc, source, chanpair, perm='list')

    dbentry = db.get(ircobj.name+channel)
    if not dbentry:
        error(irc, "No Automode access entries exist for \x02%s\x02." % channel)
        return

    else:
        # Iterate over all entries and print them. Do this in private to prevent channel
        # floods.
        reply(irc, "Showing Automode entries for \x02%s\x02:" % channel, private=True)
        for entrynum, entry in enumerate(sorted(dbentry.items()), start=1):
            mask, modes = entry
            reply(irc, "[%s] \x02%s\x02 has modes +\x02%s\x02" % (entrynum, mask, modes), private=True)
        reply(irc, "End of Automode entries list.", private=True)

modebot.add_cmd(listacc, featured=True, aliases=('listaccess',))

def save(irc, source, args):
    """takes no arguments.

    Saves the Automode database to disk."""
    permissions.check_permissions(irc, source, ['automode.savedb'])
    datastore.save()
    reply(irc, 'Done.')

modebot.add_cmd(save)

def syncacc(irc, source, args):
    """<channel/chanpair>

    Syncs Automode access lists to the channel.
    """
    try:
        chanpair = args[0]
    except IndexError:
        error(irc, "Invalid arguments given. Needs 1: channel.")
        return
    else:
        ircobj, channel = _get_channel_pair(irc, source, chanpair, perm='sync')

    log.info('(%s) %s synced modes on %s', ircobj.name, irc.get_hostmask(source), channel)
    match(ircobj, channel)

    reply(irc, 'Done.')

modebot.add_cmd(syncacc, featured=True, aliases=('sync', 'syncaccess'))

def clearacc(irc, source, args):
    """<channel>

    Removes all Automode entries for the given channel.
    """

    try:
        chanpair = args[0]
    except IndexError:
        error(irc, "Invalid arguments given. Needs 1: channel.")
        return
    else:
        ircobj, channel = _get_channel_pair(irc, source, chanpair, perm='clear')

    if db.get(ircobj.name+channel):
        del db[ircobj.name+channel]
        log.info('(%s) %s cleared modes on %s', ircobj.name, irc.get_hostmask(source), channel)
        reply(irc, "Done. Removed all Automode access entries for \x02%s\x02." % channel)
        modebot.remove_persistent_channel(ircobj, 'automode', channel)
    else:
        error(irc, "No Automode access entries exist for \x02%s\x02." % channel)

modebot.add_cmd(clearacc, aliases=('clearaccess', 'clear'), featured=True)



"""
structures.py - PyLink data structures module.

This module contains custom data structures that may be useful in various situations.
"""

import collections
import collections.abc
import json
import pickle
import os
import threading
from copy import copy, deepcopy
import string

from .log import log
from . import conf

_BLACKLISTED_COPY_TYPES = []

class KeyedDefaultdict(collections.defaultdict):
    """
    Subclass of defaultdict allowing the key to be passed to the default factory.
    """
    def __missing__(self, key):
        if self.default_factory is None:
            # If there is no default factory, just let defaultdict handle it
            super().__missing__(self, key)
        else:
            value = self[key] = self.default_factory(key)
            return value

class CopyWrapper():
    """
    Base container class implementing copy methods.
    """

    def copy(self):
        """Returns a shallow copy of this object instance."""
        return copy(self)

    def __deepcopy__(self, memo):
        """Returns a deep copy of the channel object."""
        newobj = copy(self)
        log.debug('CopyWrapper: _BLACKLISTED_COPY_TYPES = %s', _BLACKLISTED_COPY_TYPES)
        for attr, val in self.__dict__.items():
            # We can't pickle IRCNetwork, so just return a reference of it.
            if not isinstance(val, tuple(_BLACKLISTED_COPY_TYPES)):
                log.debug('CopyWrapper: copying attr %r', attr)
                setattr(newobj, attr, deepcopy(val))

        memo[id(self)] = newobj

        return newobj

    def deepcopy(self):
        """Returns a deep copy of this object instance."""
        return deepcopy(self)

class CaseInsensitiveFixedSet(collections.abc.Set, CopyWrapper):
    """
    Implements a fixed set storing items case-insensitively.
    """

    def __init__(self, *, data=None):
        if data is not None:
            assert isinstance(data, set)
            self._data = data
        else:
            self._data = set()

    @staticmethod
    def _keymangle(key):
        """Converts the given key to lowercase."""
        if isinstance(key, str):
            return key.lower()
        return key

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._data)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        return self._data.__contains__(self._keymangle(key))

    def __copy__(self):
        return self.__class__(data=self._data.copy())

class CaseInsensitiveDict(collections.abc.MutableMapping, CaseInsensitiveFixedSet):
    """
    A dictionary storing items case insensitively.
    """
    def __init__(self, *, data=None):
        if data is not None:
            assert isinstance(data, dict)
            self._data = data
        else:
            self._data = {}

    def __getitem__(self, key):
        key = self._keymangle(key)

        return self._data[key]

    def __setitem__(self, key, value):
        self._data[self._keymangle(key)] = value

    def __delitem__(self, key):
        del self._data[self._keymangle(key)]

class IRCCaseInsensitiveDict(CaseInsensitiveDict):
    """
    A dictionary storing items case insensitively, using IRC case mappings.
    """
    def __init__(self, irc, *, data=None):
        super().__init__(data=data)
        self._irc = irc

    def _keymangle(self, key):
        """Converts the given key to lowercase."""
        if isinstance(key, str):
            return self._irc.to_lower(key)
        return key

    def __copy__(self):
        return self.__class__(self._irc, data=self._data.copy())

class CaseInsensitiveSet(collections.abc.MutableSet, CaseInsensitiveFixedSet):
    """
    A mutable set storing items case insensitively.
    """

    def add(self, key):
        self._data.add(self._keymangle(key))

    def discard(self, key):
        self._data.discard(self._keymangle(key))

class IRCCaseInsensitiveSet(CaseInsensitiveSet):
    """
    A set storing items case insensitively, using IRC case mappings.
    """
    def __init__(self, irc, *, data=None):
        super().__init__(data=data)
        self._irc = irc

    def _keymangle(self, key):
        """Converts the given key to lowercase."""
        if isinstance(key, str):
            return self._irc.to_lower(key)
        return key

    def __copy__(self):
        return self.__class__(self._irc, data=self._data.copy())

class CamelCaseToSnakeCase():
    """
    Class which automatically converts missing attributes from camel case to snake case.
    """

    def __getattr__(self, attr):
        """
        Attribute fetching fallback function which normalizes camel case attributes to snake case.
        """
        assert isinstance(attr, str), "Requested attribute %r is not a string!" % attr

        normalized_attr = ''  # Start off with the first letter, which is ignored when processing
        for char in attr:
            if char in string.ascii_uppercase:
                char = '_' + char.lower()
            normalized_attr += char

        classname = self.__class__.__name__
        if normalized_attr == attr:
            # __getattr__ only fires if normal attribute fetching fails, so we can assume that
            # the attribute was tried already and failed.
            raise AttributeError('%s object has no attribute with normalized name %r' % (classname, attr))

        target = getattr(self, normalized_attr)
        log.warning('%s.%s is deprecated, considering migrating to %s.%s!', classname, attr, classname, normalized_attr)
        return target

class DataStore:
    """
    Generic database class. Plugins should use a subclass of this such as JSONDataStore or
    PickleDataStore.
    """
    def __init__(self, name, filename, save_frequency=None, default_db=None):
        self.name = name
        self.filename = filename
        self.tmp_filename = filename + '.tmp'

        log.debug('(DataStore:%s) using implementation %s', self.name, self.__class__.__name__)
        log.debug('(DataStore:%s) database path set to %s', self.name, self.filename)

        self.save_frequency = save_frequency or conf.conf['pylink'].get('save_delay', 300)
        log.debug('(DataStore:%s) saving every %s seconds', self.name, self.save_frequency)

        if default_db is not None:
            self.store = default_db
        else:
            self.store = {}
        self.store_lock = threading.Lock()
        self.exportdb_timer = None

        self.load()

        if self.save_frequency > 0:
            # If autosaving is enabled, start the save_callback loop.
            self.save_callback(starting=True)

    def load(self):
        """
        DataStore load stub. Database implementations should subclass DataStore
        and implement this.
        """
        raise NotImplementedError

    def save_callback(self, starting=False):
        """Start the DB save loop."""
        # don't actually save the first time
        if not starting:
            self.save()

        # schedule saving in a loop.
        self.exportdb_timer = threading.Timer(self.save_frequency, self.save_callback)
        self.exportdb_timer.name = 'DataStore {} save_callback loop'.format(self.name)
        self.exportdb_timer.start()

    def save(self):
        """
        DataStore save stub. Database implementations should subclass DataStore
        and implement this.
        """
        raise NotImplementedError

    def die(self):
        """
        Saves the database and stops any save loops.
        """
        if self.exportdb_timer:
            self.exportdb_timer.cancel()

        self.save()

class JSONDataStore(DataStore):
    def load(self):
        """Loads the database given via JSON."""
        with self.store_lock:
            try:
                with open(self.filename, "r") as f:
                    self.store.clear()
                    self.store.update(json.load(f))
            except (ValueError, IOError, OSError):
                log.info("(DataStore:%s) failed to load database %s; creating a new one in "
                         "memory", self.name, self.filename)

    def save(self):
        """Saves the database given via JSON."""
        with self.store_lock:
            with open(self.tmp_filename, 'w') as f:
                # Pretty print the JSON output for better readability.
                json.dump(self.store, f, indent=4)

                os.rename(self.tmp_filename, self.filename)

class PickleDataStore(DataStore):
    def load(self):
        """Loads the database given via pickle."""
        with self.store_lock:
            try:
                with open(self.filename, "rb") as f:
                    self.store.clear()
                    self.store.update(pickle.load(f))
            except (ValueError, IOError, OSError):
                log.info("(DataStore:%s) failed to load database %s; creating a new one in "
                         "memory", self.name, self.filename)

    def save(self):
        """Saves the database given via pickle."""
        with self.store_lock:
            with open(self.tmp_filename, 'wb') as f:
                # Force protocol version 4 as that is the lowest Python 3.4 supports.
                pickle.dump(self.store, f, protocol=4)

                os.rename(self.tmp_filename, self.filename)



"""
ngircd.py: PyLink protocol module for ngIRCd.
"""
##
# Server protocol docs for ngIRCd can be found at:
#     https://github.com/ngircd/ngircd/blob/master/doc/Protocol.txt
# and https://tools.ietf.org/html/rfc2813
##

import time
import re

from pylinkirc import utils, conf, __version__
from pylinkirc.classes import *
from pylinkirc.log import log
from pylinkirc.protocols.ircs2s_common import *

class NgIRCdProtocol(IRCS2SProtocol):
    def __init__(self, irc):
        super().__init__(irc)

        self.conf_keys -= {'sid', 'sidrange'}
        self.casemapping = 'ascii'  # This is the default; it's actually set on server negotiation
        self.hook_map = {'NJOIN': 'JOIN'}

        # Track whether we've received end-of-burst from the uplink.
        self.has_eob = False

        self._caps = {}
        self._use_builtin_005_handling = True

        # ngIRCd has no TS tracking.
        self.protocol_caps.discard('has-ts')

        # Slash in nicks is problematic; while it works for basic things like JOIN and messages,
        # attempts to set user modes fail.
        self.protocol_caps |= {'slash-in-hosts', 'underscore-in-hosts'}

    ### Commands

    def post_connect(self):
        self.send('PASS %s 0210-IRC+ PyLink|%s:CHLMoX' % (self.serverdata['sendpass'], __version__))

        # Note: RFC 2813 mandates another server token value after the hopcount (1), but ngIRCd
        # doesn't follow that behaviour per https://github.com/ngircd/ngircd/issues/224
        self.send("SERVER %s 1 :%s" % (self.serverdata['hostname'],
                                       self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']))

        self._uidgen = PUIDGenerator('PUID')

        # The first "SID" this generator should return is 2, because server token 1 is implied to be
        # the main PyLink server. RFC2813 has no official definition of SIDs, but rather uses
        # integer tokens in the SERVER and NICK (user introduction) commands to keep track of which
        # user exists on which server. Why did they do it this way? Who knows!
        self._sidgen = PUIDGenerator('PSID', start=1)
        self.sid = self._sidgen.next_sid(prefix=self.serverdata['hostname'])

        self._caps.clear()

        self.cmodes.update({
            'banexception': 'e', 'invex': 'I', 'regmoderated': 'M', 'nonick': 'N',
            'operonly': 'O', 'permanent': 'P', 'nokick': 'Q', 'registered': 'r',
            'regonly': 'R', 'noinvite': 'V', 'sslonly': 'z'
        })

        self.umodes.update({
            'away': 'a', 'deaf': 'b', 'bot': 'B', 'sno_clientconnections': 'c',
            'deaf_commonchan': 'C', 'floodexempt': 'f', 'hidechans': 'I',
            'servprotect': 'q', 'restricted': 'r', 'registered': 'R', 'cloak': 'x'
        })

    def spawn_client(self, nick, ident='null', host='null', realhost=None, modes=set(),
            server=None, ip='0.0.0.0', realname=None, ts=None, opertype='IRC Operator',
            manipulatable=False):
        """
        Spawns a new client with the given options.

        Note: No nick collision / valid nickname checks are done here; it is
        up to plugins to make sure they don't introduce anything invalid.

        Note 2: IP and realhost are ignored because ngIRCd does not send them.
        """
        server = server or self.sid
        assert '@' in server, "Need PSID for spawn_client, not pure server name!"
        if not self.is_internal_server(server):
            raise ValueError('Server %r is not a PyLink server!' % server)

        realname = realname or conf.conf['pylink']['realname']

        uid = self._uidgen.next_uid(prefix=nick)
        userobj = self.users[uid] = User(self, nick, ts or int(time.time()), uid, server,
                                         ident=ident, host=host, realname=realname,
                                         manipulatable=manipulatable, opertype=opertype,
                                         realhost=host)

        self.apply_modes(uid, modes)
        self.servers[server].users.add(uid)

        # Grab our server token; this is used instead of server name to denote where the client is.
        server_token = server.rsplit('@')[-1]
        # <- :ngircd.midnight.local NICK GL 1 ~gl localhost 1 +io :realname
        self._send_with_prefix(server, 'NICK %s %s %s %s %s %s :%s' % (nick, self.servers[server].hopcount,
                               ident, host, server_token, self.join_modes(modes), realname))
        return userobj

    def spawn_server(self, name, sid=None, uplink=None, desc=None):
        """
        Spawns a server off a PyLink server.

        * desc (server description) defaults to the one in the config.
        * uplink defaults to the main PyLink server.
        * SID is set equal to the server name for ngIRCd, as server UIDs are not used.
        """
        uplink = uplink or self.sid
        assert uplink in self.servers, "Unknown uplink %r?" % uplink
        name = name.lower()
        sid = self._sidgen.next_sid(prefix=name)

        desc = desc or self.serverdata.get('serverdesc') or conf.conf['pylink']['serverdesc']

        if sid in self.servers:
            raise ValueError('A server named %r already exists!' % sid)

        if not self.is_internal_server(uplink):
            raise ValueError('Server %r is not a PyLink server!' % uplink)

        if not self.is_server_name(name):
            raise ValueError('Invalid server name %r' % name)

        # https://tools.ietf.org/html/rfc2813#section-4.1.2
        # We need to store a server token to introduce clients on the right server. Since this is just
        # a number, we can simply use the counter in our PSID generator for this.
        server_token = sid.rsplit('@')[-1]
        self.servers[sid] = Server(self, uplink, name, internal=True, desc=desc)
        self._send_with_prefix(uplink, 'SERVER %s %s %s :%s' % (name, self.servers[sid].hopcount, server_token, desc))
        return sid

    def away(self, source, text):
        """Sends an AWAY message from a PyLink client. If the text is empty, away status is unset."""
        if not self.is_internal_client(source):
            raise LookupError('No such PyLink client exists.')

        # Away status is denoted on ngIRCd with umode +a.
        modes = self.users[source].modes
        if text and (('a', None) not in modes):
            # Set umode +a if it isn't already set
            self.mode(source, source, [('+a', None)])
        elif ('a', None) in modes:
            # Ditto, only unset the mode if it *was* set.
            self.mode(source, source, [('-a', None)])
        self.users[source].away = text

    def join(self, client, channel):

        if not self.is_internal_client(client):
            raise LookupError('No such PyLink client exists.')

        self._send_with_prefix(client, "JOIN %s" % channel)
        self._channels[channel].users.add(client)
        self.users[client].channels.add(channel)

    def kill(self, source, target, reason):
        """Sends a kill from a PyLink client/server."""
        if (not self.is_internal_client(source)) and \
                (not self.is_internal_server(source)):
            raise LookupError('No such PyLink client/server exists.')

        # Follow ngIRCd's formatting of the kill messages for the most part
        self._send_with_prefix(source, 'KILL %s :KILLed by %s: %s' % (self._expandPUID(target),
                               self.get_friendly_name(source), reason))

        # Implicitly remove our own client if one was the target.
        if self.is_internal_client(target):
            self._remove_client(target)

    def knock(self, numeric, target, text):
        raise NotImplementedError('KNOCK is not supported on ngIRCd.')

    def mode(self, source, target, modes, ts=None):
        """Sends mode changes from a PyLink client/server. The TS option is not used on ngIRCd."""

        if (not self.is_internal_client(source)) and \
                (not self.is_internal_server(source)):
            raise LookupError('No such PyLink client/server %r exists' % source)

        self.apply_modes(target, modes)
        modes = list(modes)  # Work around TypeError in the expand PUID section

        if self.is_channel(target):
            msgprefix = ':%s MODE %s ' % (self._expandPUID(source), target)
            bufsize = self.S2S_BUFSIZE - len(msgprefix)

            # Expand PUIDs when sending outgoing prefix modes.
            for idx, mode in enumerate(modes):
                if mode[0][-1] in self.prefixmodes:
                    log.debug('(%s) mode: expanding PUID of mode %s', self.name, str(mode))
                    modes[idx] = (mode[0], self._expandPUID(mode[1]))

            for modestr in self.wrap_modes(modes, bufsize, max_modes_per_msg=12):
                self.send(msgprefix + modestr)
        else:
            joinedmodes = self.join_modes(modes)
            self._send_with_prefix(source, 'MODE %s %s' % (self._expandPUID(target), joinedmodes))

    def nick(self, source, newnick):
        """Changes the nick of a PyLink client."""
        if not self.is_internal_client(source):
            raise LookupError('No such PyLink client exists.')

        self._send_with_prefix(source, 'NICK %s' % newnick)

        self.users[source].nick = newnick

        # Update the nick TS for consistency with other protocols (it isn't actually used in S2S)
        self.users[source].ts = int(time.time())

    def sjoin(self, server, channel, users, ts=None, modes=set()):
        """Sends an SJOIN for a group of users to a channel.

        The sender should always be a Server ID (SID). TS is optional, and defaults
        to the one we've stored in the channel state if not given.
        <users> is a list of (prefix mode, UID) pairs:

        Example uses:
            sjoin('100', '#test', [('', 'user0@0'), ('o', user1@1'), ('v', 'someone@2')])
            sjoin(self.sid, '#test', [('o', self.pseudoclient.uid)])
        """

        server = server or self.sid
        if not server:
            raise LookupError('No such PyLink client exists.')
        log.debug('(%s) sjoin: got %r for users', self.name, users)

        njoin_prefix = ':%s NJOIN %s :' % (self._expandPUID(server), channel)
        # Format the user list into strings such as @user1, +user2, user3, etc.
        nicks_to_send = []
        for userpair in users:
            prefixes, uid = userpair

            if uid not in self.users:
                log.warning('(%s) Trying to NJOIN missing user %s?', self.name, uid)
                continue
            elif uid in self._channels[channel].users:
                # Don't rejoin users already in the channel, this causes errors with ngIRCd.
                continue

            self._channels[channel].users.add(uid)
            self.users[uid].channels.add(channel)

            self.apply_modes(channel, (('+%s' % prefix, uid) for prefix in userpair[0]))

            nicks_to_send.append(''.join(self.prefixmodes[modechar] for modechar in userpair[0]) + \
                                 self._expandPUID(userpair[1]))

        if nicks_to_send:
            # Use 13 args max per line: this is equal to the max of 15 minus the command name and target channel.
            for message in utils.wrap_arguments(njoin_prefix, nicks_to_send, self.S2S_BUFSIZE, separator=',', max_args_per_line=13):
                self.send(message)

        if modes:
            # Burst modes separately if there are any.
            log.debug("(%s) sjoin: bursting modes %r for channel %r now", self.name, modes, channel)
            self.mode(server, channel, modes)

    def set_server_ban(self, source, duration, user='*', host='*', reason='User banned'):
        """
        Sets a server ban.
        """
        # <- :GL GLINE *!*@bad.user 3d :test
        assert not (user == host == '*'), "Refusing to set ridiculous ban on *@*"
        self._send_with_prefix(source, 'GLINE *!%s@%s %s :%s' % (user, host, duration, reason))

    def update_client(self, target, field, text):
        """Updates the ident, host, or realname of any connected client."""
        field = field.upper()

        if field not in ('IDENT', 'HOST', 'REALNAME', 'GECOS'):
            raise NotImplementedError("Changing field %r of a client is "
                                      "unsupported by this protocol." % field)

        real_target = self._expandPUID(target)
        if field == 'IDENT':
            self.users[target].ident = text
            self._send_with_prefix(self.sid, 'METADATA %s user :%s' % (real_target, text))

            if not self.is_internal_client(target):
                # If the target wasn't one of our clients, send a hook payload for other plugins to listen to.
                self.call_hooks([self.sid, 'CHGIDENT', {'target': target, 'newident': text}])

        elif field == 'HOST':
            self.users[target].host = text

            if self.is_internal_client(target):
                # For our own clients, replace the real host.
                self._send_with_prefix(self.sid, 'METADATA %s host :%s' % (real_target, text))
            else:
                # For others, update the cloaked host and force a umode +x.
                self._send_with_prefix(self.sid, 'METADATA %s cloakhost :%s' % (real_target, text))

                if ('x', None) not in self.users[target].modes:
                    log.debug('(%s) Forcing umode +x on %r as part of cloak setting', self.name, target)
                    self.mode(self.sid, target, [('+x', None)])

                self.call_hooks([self.sid, 'CHGHOST', {'target': target, 'newhost': text}])

        elif field in ('REALNAME', 'GECOS'):
            self.users[target].realname = text
            self._send_with_prefix(self.sid, 'METADATA %s info :%s' % (real_target, text))

            if not self.is_internal_client(target):
                self.call_hooks([self.sid, 'CHGNAME', {'target': target, 'newgecos': text}])

    ### Handlers

    def handle_376(self, source, command, args):
        # 376 is used to denote end of server negotiation - we send our info back at this point.
        # <- :ngircd.midnight.local 005 pylink-devel.int NETWORK=ngircd-test :is my network name
        # <- :ngircd.midnight.local 005 pylink-devel.int RFC2812 IRCD=ngIRCd CHARSET=UTF-8 CASEMAPPING=ascii PREFIX=(qaohv)~&@%+ CHANTYPES=#&+ CHANMODES=beI,k,l,imMnOPQRstVz CHANLIMIT=#&+:10 :are supported on this server
        # <- :ngircd.midnight.local 005 pylink-devel.int CHANNELLEN=50 NICKLEN=21 TOPICLEN=490 AWAYLEN=127 KICKLEN=400 MODES=5 MAXLIST=beI:50 EXCEPTS=e INVEX=I PENALTY :are supported on this server
        def f(numeric, msg):
            self._send_with_prefix(self.sid, '%s %s %s' % (numeric, self.uplink, msg))
        f('005', 'NETWORK=%s :is my network name' % self.get_full_network_name())
        f('005', 'RFC2812 IRCD=PyLink CHARSET=UTF-8 CASEMAPPING=%s PREFIX=%s CHANTYPES=# '
          'CHANMODES=%s,%s,%s,%s :are supported on this server' % (self.casemapping, self._caps['PREFIX'],
          self.cmodes['*A'], self.cmodes['*B'], self.cmodes['*C'], self.cmodes['*D']))
        f('005', 'CHANNELLEN NICKLEN=%s EXCEPTS=E INVEX=I :are supported on this server' % self.maxnicklen)

        # 376 (end of MOTD) marks the end of extended server negotiation per
        # https://github.com/ngircd/ngircd/blob/master/doc/Protocol.txt#L103-L112
        f('376', ":End of server negotiation, happy PyLink'ing!")

    def handle_chaninfo(self, source, command, args):
        # https://github.com/ngircd/ngircd/blob/722afc1b810cef74dbd2738d71866176fd974ec2/doc/Protocol.txt#L146-L159
        # CHANINFO has 3 styles depending on the amount of information applicable to a channel:
        #    CHANINFO <channel> +<modes>
        #    CHANINFO <channel> +<modes> <topic>
        #    CHANINFO <channel> +<modes> <key> <limit> <topic>
        # If there is no key, the key is "*". If there is no limit, the limit is "0".

        channel = args[0]
        # Get rid of +l and +k in the initial parsing; we handle that later by looking at the CHANINFO arguments
        modes = self.parse_modes(channel, args[1].replace('l', '').replace('k', ''))

        if len(args) >= 3:
            topic = args[-1]
            if topic:
                log.debug('(%s) handle_chaninfo: setting topic for %s to %r', self.name, channel, topic)
                self._channels[channel].topic = topic
                self._channels[channel].topicset = True

        if len(args) >= 5:
            key = args[2]
            limit = args[3]
            if key != '*':
                modes.append(('+k', key))
            if limit != '0':
                modes.append(('+l', limit))

        self.apply_modes(channel, modes)

    def handle_join(self, source, command, args):
        # RFC 2813 is odd to say the least... https://tools.ietf.org/html/rfc2813#section-4.2.1
        # Basically, we expect messages of the forms:
        # <- :GL JOIN #test\x07o
        # <- :GL JOIN #moretest
        for chanpair in args[0].split(','):
            # Normalize channel case.
            try:
                channel, status = chanpair.split('\x07', 1)
                if status in 'ov':
                    self.apply_modes(channel, [('+' + status, source)])
            except ValueError:
                channel = chanpair

            c = self._channels[channel]

            self.users[source].channels.add(channel)
            self._channels[channel].users.add(source)

            # Call hooks manually, because one JOIN command have multiple channels.
            self.call_hooks([source, command, {'channel': channel, 'users': [source], 'modes': c.modes}])

    def handle_kill(self, source, command, args):
        """Handles incoming KILLs."""

        # ngIRCd sends QUIT after KILL for its own clients, so we shouldn't process this by itself
        # unless we're the target.
        killed = self._get_UID(args[0])
        if self.is_internal_client(killed):
            return super().handle_kill(source, command, args)
        else:
            log.debug("(%s) Ignoring KILL to %r as it isn't meant for us; we should see a QUIT soon",
                      self.name, killed)

    def _check_cloak_change(self, target):
        u = self.users[target]
        old_host = u.host

        if ('x', None) in u.modes and u.cloaked_host:
            u.host = u.cloaked_host
        elif u.realhost:
            u.host = u.realhost

        # Something changed, so send a CHGHOST hook
        if old_host != u.host:
            self.call_hooks([target, 'CHGHOST', {'target': target, 'newhost': u.host}])

    def handle_metadata(self, source, command, args):
        """Handles various user metadata for ngIRCd (cloaked host, account name, etc.)"""
        # <- :ngircd.midnight.local METADATA GL cloakhost :hidden-3a2a739e.ngircd.midnight.local
        target = self._get_UID(args[0])

        if target not in self.users:
            log.warning("(%s) Ignoring METADATA to missing user %r?", self.name, target)
            return

        datatype = args[1]
        u = self.users[target]

        if datatype == 'cloakhost':  # Set cloaked host
            u.cloaked_host = args[-1]
            self._check_cloak_change(target)

        elif datatype == 'host':  # Host changing. This actually sets the "real host" that ngIRCd stores
            u.realhost = args[-1]
            self._check_cloak_change(target)

        elif datatype == 'user':  # Ident changing
            u.ident = args[-1]
            self.call_hooks([target, 'CHGIDENT', {'target': target, 'newident': args[-1]}])

        elif datatype == 'info':  # Realname changing
            u.realname = args[-1]
            self.call_hooks([target, 'CHGNAME', {'target': target, 'newgecos': args[-1]}])

        elif datatype == 'accountname':  # Services account
            self.call_hooks([target, 'CLIENT_SERVICES_LOGIN', {'text': args[-1]}])

    def handle_nick(self, source, command, args):
        """
        Handles the NICK command, used for server introductions and nick changes.
        """
        if len(args) >= 2:
            # User introduction:
            # <- :ngircd.midnight.local NICK GL 1 ~gl localhost 1 +io :realname
            nick = args[0]
            assert source in self.servers, "Server %r tried to introduce nick %r but isn't in the servers index?" % (source, nick)
            self._check_nick_collision(nick)

            ident = args[2]
            host = args[3]
            uid = self._uidgen.next_uid(prefix=nick)
            realname = args[-1]

            ts = int(time.time())
            self.users[uid] = User(self, nick, ts, uid, source, ident=ident, host=host,
                                   realname=realname, realhost=host)
            parsedmodes = self.parse_modes(uid, [args[5]])
            self.apply_modes(uid, parsedmodes)

            # Add the nick to the list of users on its server; this is used for SQUIT tracking
            self.servers[source].users.add(uid)

            # Check away status and cloaked host changes
            self._check_umode_away_change(uid)
            self._check_cloak_change(uid)

            return {'uid': uid, 'ts': ts, 'nick': nick, 'realhost': host, 'host': host, 'ident': ident,
                    'parse_as': 'UID', 'ip': '0.0.0.0'}
        else:
            # Nick changes:
            # <- :GL NICK :GL_
            oldnick = self.users[source].nick
            newnick = self.users[source].nick = args[0]
            return {'newnick': newnick, 'oldnick': oldnick}

    def handle_njoin(self, source, command, args):
        # <- :ngircd.midnight.local NJOIN #test :tester,@%GL

        channel = args[0]
        chandata = self._channels[channel].deepcopy()
        namelist = []

        # Reverse the modechar->modeprefix mapping for quicker lookup
        prefixchars = {v: k for k, v in self.prefixmodes.items()}
        for userpair in args[1].split(','):
            # Some regex magic to split the prefix from the nick.
            r = re.search(r'([%s]*)(.*)' % ''.join(self.prefixmodes.values()), userpair)
            user = self._get_UID(r.group(2))
            modeprefix = r.group(1)

            if modeprefix:
                modes = {('+%s' % prefixchars[mode], user) for mode in modeprefix}
                self.apply_modes(channel, modes)
            namelist.append(user)

            # Final bits of state tracking. (I hate having to do this everywhere...)
            self.users[user].channels.add(channel)
            self._channels[channel].users.add(user)

        return {'channel': channel, 'users': namelist, 'modes': [], 'channeldata': chandata}

    def handle_pass(self, source, command, args):
        """
        Handles phase one of the ngIRCd login process (password auth and version info).
        """
        # PASS is step one of server introduction, and is used to send the server info and password.
        # <- :ngircd.midnight.local PASS xyzpassword 0210-IRC+ ngIRCd|24~3-gbc728f92:CHLMSXZ PZ
        recvpass = args[0]
        if recvpass != self.serverdata['recvpass']:
            raise ProtocolError("RECVPASS from uplink does not match configuration!")

        assert 'IRC+' in args[1], "Linking to non-ngIRCd server using this protocol module is not supported"

    def handle_ping(self, source, command, args):
        """
        Handles incoming PINGs (and implicit end of burst).
        """
        self._send_with_prefix(self.sid, 'PONG %s :%s' % (self._expandPUID(self.sid), args[-1]), queue=False)

        if not self.servers[source].has_eob:
            # Treat the first PING we receive as end of burst.
            self.servers[source].has_eob = True

            if source == self.uplink:
                self.connected.set()

            # Return the endburst hook.
            return {'parse_as': 'ENDBURST'}

    def handle_server(self, source, command, args):
        """
        Handles the SERVER command.
        """
        # <- :ngircd.midnight.local SERVER ngircd.midnight.local 1 :ngIRCd dev server
        servername = args[0].lower()
        serverdesc = args[-1]

        # The uplink should be set to None for the uplink; otherwise, set it equal to the sender server.
        self.servers[servername] = Server(self, source if source != servername else None, servername, desc=serverdesc)

        if self.uplink is None:
            self.uplink = servername
            log.debug('(%s) Got %s as uplink', self.name, servername)
        else:
            # Only send the SERVER hook if this isn't the initial connection.
            return {'name': servername, 'sid': None, 'text': serverdesc}

Class = NgIRCdProtocol



"""
Changehost plugin - automatically changes the hostname of matching users.
"""
from pylinkirc import utils, world, conf
from pylinkirc.log import log
from pylinkirc.coremods import permissions

import string

# ircmatch library from https://github.com/mammon-ircd/ircmatch
# (pip install ircmatch)
import ircmatch

# Characters allowed in a hostname.
allowed_chars = string.ascii_letters + '-./:' + string.digits

def _changehost(irc, target, args):
    changehost_conf = conf.conf.get("changehost")

    if target not in irc.users:
        return
    elif irc.is_internal_client(target):
        log.debug('(%s) Skipping changehost on internal client %s', irc.name, target)
        return

    if not changehost_conf:
        log.warning("(%s) Missing 'changehost:' configuration block; "
                    "Changehost will not function correctly!", irc.name)
        return
    elif irc.name not in changehost_conf.get('enabled_nets'):
        # We're not enabled on the network, break.
        return

    match_ip = changehost_conf.get('match_ip', False)
    match_realhosts = changehost_conf.get('match_realhosts', False)

    changehost_hosts = changehost_conf.get('hosts')
    if not changehost_hosts:
        log.warning("(%s) No hosts were defined in changehost::hosts; "
                    "Changehost will not function correctly!", irc.name)
        return

    args = args.copy()
    if not changehost_conf.get('force_host_expansion'):
       del args['host']

    log.debug('(%s) Changehost args: %s', irc.name, args)

    for host_glob, host_template in changehost_hosts.items():
        log.debug('(%s) Changehost: checking mask %s', irc.name, host_glob)
        if irc.match_host(host_glob, target, ip=match_ip, realhost=match_realhosts):
            log.debug('(%s) Changehost matched mask %s', irc.name, host_glob)
            # This uses template strings for simple substitution:
            # https://docs.python.org/3/library/string.html#template-strings
            template = string.Template(host_template)

            # Substitute using the fields provided the hook data. This means
            # that the following variables are available for substitution:
            # $uid, $ts, $nick, $realhost, $ident, and $ip.

            # $host is explicitly forbidden by default because it can cause
            # recursive loops when IP or real host masks are used to match a
            # target. vHost updates do not affect these fields, so any further
            # execution of 'applyhosts' will cause $host to expand again to
            # the user's new host, causing the vHost to grow rapidly in size.
            # That said, it is possible to get away with this expansion if
            # you're careful with what you're doing, and that is why this
            # hidden option exists. -GLolol

            try:
                new_host = template.substitute(args)
            except KeyError as e:
                log.warning('(%s) Bad expansion %s in template %s' % (irc.name, e, host_template))
                continue

            # Replace characters that are not allowed in hosts with "-".
            for char in new_host:
                if char not in allowed_chars:
                    new_host = new_host.replace(char, '-')

            irc.update_client(target, 'HOST', new_host)

            # Only operate on the first match.
            break

def handle_uid(irc, sender, command, args):
    """
    Changehost listener for new connections.
    """

    target = args['uid']
    _changehost(irc, target, args)

utils.add_hook(handle_uid, 'UID')

def handle_chghost(irc, sender, command, args):
    """
    Handles incoming CHGHOST requests for optional host-change enforcement.
    """
    changehost_conf = conf.conf.get("changehost")
    if not changehost_conf:
        return

    target = args['target']

    if (not irc.is_internal_client(sender)) and (not irc.is_internal_server(sender)):
        if irc.name in changehost_conf.get('enforced_nets', []):
            log.debug('(%s) Enforce for network is on, re-checking host for target %s/%s',
                      irc.name, target, irc.get_friendly_name(target))

            for ex in changehost_conf.get("enforce_exceptions", []):
                if irc.match_host(ex, target):
                    log.debug('(%s) Skipping host change for target %s; they are exempted by mask %s',
                              irc.name, target, ex)
                    return

            userobj = irc.users.get(target)
            if userobj:
                _changehost(irc, target, userobj.get_fields())

utils.add_hook(handle_chghost, 'CHGHOST')

@utils.add_cmd
def applyhosts(irc, sender, args):
    """[<network>]

    Applies all configured hosts for users on the given network, or the current network if none is specified."""

    permissions.check_permissions(irc, sender, ['changehost.applyhosts'])

    try:  # Try to get network from the command line.
        network = world.networkobjects[args[0]]
    except IndexError:  # No network was given
        network = irc
    except KeyError:  # Unknown network
        irc.error("Unknown network '%s'." % network)
        return

    for user, userdata in network.users.copy().items():
        _changehost(network, user, userdata.__dict__)

    irc.reply("Done.")



"""
raw.py: Provides a 'raw' command for sending raw text to IRC.
"""
from pylinkirc.log import log
from pylinkirc.coremods import permissions
from pylinkirc import utils

@utils.add_cmd
def raw(irc, source, args):
    """<text>

    Sends raw text to the IRC server.

    This command is not officially supported on non-Clientbot networks, where it
    requires a separate permission."""

    if irc.protoname == 'clientbot':
        # exec.raw is included for backwards compatibility with PyLink 1.x
        perms = ['raw.raw', 'exec.raw']
    else:
        perms = ['raw.raw.unsupported_network']
    permissions.check_permissions(irc, source, perms)

    args = ' '.join(args)
    if not args.strip():
        irc.reply('No text entered!')
        return

    # Note: This is loglevel debug so that we don't risk leaking things like
    # NickServ passwords on Clientbot networks.
    log.debug('(%s) Sending raw text %r to IRC for %s', irc.name, args,
              irc.get_hostmask(source))
    irc.send(args)

    irc.reply("Done.")



"""
stats.py: Simple statistics for PyLink IRC Services.
"""
import time
import datetime

from pylinkirc import utils, world, conf
from pylinkirc.log import log
from pylinkirc.coremods import permissions

def timediff(before, now):
    """
    Returns the time difference between "before" and "now" as a formatted string.
    """
    td = datetime.timedelta(seconds=now-before)
    days = td.days

    hours, leftover = divmod(td.seconds, 3600)
    minutes, seconds = divmod(leftover, 60)

    # XXX: I would make this more configurable but it's a lot of work for little gain,
    # since there's no strftime for time differences.
    return '%d day%s, %02d:%02d:%02d' % (td.days, 's' if td.days != 1 else '',
                                         hours, minutes, seconds)

# From RFC 2822: https://tools.ietf.org/html/rfc2822.html#section-3.3
DEFAULT_TIME_FORMAT = "%a, %d %b %Y %H:%M:%S +0000"

@utils.add_cmd
def uptime(irc, source, args):
    """[<network> / --all]

    Returns the uptime for PyLink and the given network's connection (or the current network if not specified).
    The --all argument can also be given to show the uptime for all networks."""
    permissions.check_permissions(irc, source, ['stats.uptime'])

    try:
        network = args[0]
    except IndexError:
        network = irc.name

    if network == '--all':  # XXX: we really need smart argument parsing some time
        # Filter by all connected networks.
        ircobjs = {k:v for k,v in world.networkobjects.items() if v.connected.is_set()}
    else:
        try:
            ircobjs = {network: world.networkobjects[network]}
        except KeyError:
            irc.error("No such network %r." % network)
            return
        if not world.networkobjects[network].connected.is_set():
            irc.error("Network %s is not connected." % network)
            return

    current_time = int(time.time())
    time_format = conf.conf.get('stats', {}).get('time_format', DEFAULT_TIME_FORMAT)

    irc.reply("PyLink uptime: \x02%s\x02 (started on %s)" %
              (timediff(world.start_ts, current_time),
               time.strftime(time_format, time.gmtime(world.start_ts))
              )
             )

    for network, ircobj in sorted(ircobjs.items()):
        irc.reply("Connected to %s: \x02%s\x02 (connected on %s)" %
                  (network,
                   timediff(ircobj.start_ts, current_time),
                   time.strftime(time_format, time.gmtime(ircobj.start_ts))
                  )
                 )

def handle_stats(irc, source, command, args):
    """/STATS handler. Currently supports the following:

    c - link blocks
    o - oper blocks (accounts)
    u - shows uptime
    """

    stats_type = args['stats_type'][0].lower()  # stats_type shouldn't be more than 1 char anyways

    perms = ['stats.%s' % stats_type]

    if stats_type == 'u':
        perms.append('stats.uptime')  # Consistency

    try:
        permissions.check_permissions(irc, source, perms)
    except utils.NotAuthorizedError as e:
        # Note, no irc.error() because this is not a command, but a handler
        irc.msg(source, 'Error: %s' % e, notice=True)
        return

    log.info('(%s) /STATS %s requested by %s', irc.name, stats_type, irc.get_hostmask(source))

    def _num(num, text):
        irc.numeric(args['target'], num, source, text)

    if stats_type == 'c':
        # 213/RPL_STATSCLINE: "C <host> * <name> <port> <class>"
        for netname, serverdata in sorted(conf.conf['servers'].items()):
            # We're cramming as much as we can into the class field...
            _num(213, "C %s * %s %s [%s:%s:%s]" %
                 (serverdata.get('ip', '0.0.0.0'),
                  netname,
                  serverdata.get('port', 0),
                  serverdata['protocol'],
                  'ssl' if serverdata.get('ssl') else 'no-ssl',
                  serverdata.get('encoding', 'utf-8'))
                 )
    elif stats_type == 'o':
        # 243/RPL_STATSOLINE: "O <hostmask> * <nick> [:<info>]"
        # New style accounts only!
        for accountname, accountdata in conf.conf['login'].get('accounts', {}).items():
            _num(243, "O %s * %s :network_filter:%s require_oper:%s" %
                 (' '.join(accountdata.get('hosts', [])) or '*@*',
                  accountname,
                  ','.join(accountdata.get('networks', [])) or '*',
                  bool(accountdata.get('require_oper'))
                 )
                )

    elif stats_type == 'u':
        # 242/RPL_STATSUPTIME: ":Server Up <days> days <hours>:<minutes>:<seconds>"
        _num(242, ':Server Up %s' % timediff(world.start_ts, int(time.time())))

    else:
        log.info('(%s) Unknown /STATS type %r requested by %s', irc.name, stats_type, irc.get_hostmask(source))
    _num(219, "%s :End of /STATS report" % stats_type)
utils.add_hook(handle_stats, 'STATS')
